from typing import Optional
from langchain_core.pydantic_v1 import BaseModel
from .. import config
from . import langchain_helpers

class ExtractorLLM:
    def __init__(self):
        """Initialize the Extractor LLM with a default model."""
        self.extractor_llm = langchain_helpers.create_llm(
            chat_model=config.CHAT_MODEL,
            model_name=config.MODEL_NAME,
            temperature=config.TEMPERATURE,
            cache=True,
        )

    def extract_from_input(self, pydantic_object, input: str, **chain_kwargs) -> dict:
        """Extract structured data from input using the specified pydantic object."""
        try:
            model_with_structure = self.extractor_llm.with_structured_output(
                pydantic_object
            )
            return model_with_structure.invoke(input).dict()
        except Exception as e:
            print("Encountered exception during parsing input. See below:")
            print(e)



import unittest
from langchain.schema import HumanMessage, SystemMessage
from langchain.prompts import ChatPromptTemplate, HumanMessagePromptTemplate
from ..prompts.prompts import Prompts

PROMPT_GROUPS = ["IMPROVER", "SECTION_HIGHLIGHTER", "SKILLS_MATCHER", "OBJECTIVE_WRITER"]

PROMPT_DESCRIPTIONS = [
    "JOB_DESCRIPTION",
    "JOB_SKILLS",
    "RESUME_IMPROVEMENTS",
    "RESUME_IMPROVER_OUTPUT",
    "RESUME_SECTION_HIGHLIGHT",
    "RESUME_SECTION_HIGHLIGHTER_OUTPUT",
    "RESUME_SKILLS",
    "RESUME_SKILLS_MATCHER_OUTPUT",
    "RESUME_OBJECTIVE_OUTPUT",
]


class TestPrompts(unittest.TestCase):
    def setUp(self):
        Prompts.initialize()

    def test_load_prompts(self):
        for prompt_group in PROMPT_GROUPS:
            self.assertIn(prompt_group, Prompts.lookup)
            self.assertIsInstance(Prompts.lookup[prompt_group], list)
            self.assertGreater(len(Prompts.lookup[prompt_group]), 0)

    def test_load_descriptions(self):
        for description in PROMPT_DESCRIPTIONS:
            self.assertIn(description, Prompts.descriptions)
            self.assertIsInstance(Prompts.descriptions[description], dict)
            self.assertGreater(len(Prompts.descriptions[description]), 0)


if __name__ == "__main__":
    unittest.main()



import unittest
from ..config import config


class TestConfig(unittest.TestCase):
    def test_logger_initialization(self):
        self.assertIsNotNone(config.logger)

    def test_project_paths(self):
        self.assertTrue(config.PROJECT_PATH)
        self.assertTrue(config.DATA_PATH)
        self.assertTrue(config.RESOURCES_PATH)
        self.assertTrue(config.DEFAULT_RESUME_PATH)

    def test_requests_headers(self):
        self.assertTrue(
            any(key.lower() == "user-agent".lower() for key in config.REQUESTS_HEADERS)
        )

    def test_model_configuration_types(self):
        self.assertIsInstance(config.CHAT_MODEL, type)
        self.assertIsInstance(config.MODEL_NAME, str)
        self.assertIsInstance(config.TEMPERATURE, float)

    def test_openai_api_key(self):
        self.assertIn("OPENAI_API_KEY", config.os.environ)


if __name__ == "__main__":
    unittest.main()



from langchain_core.pydantic_v1 import BaseModel, Field
from typing import List
from ..prompts.prompts import Prompts

Prompts.initialize()

class ResumeSectionHighlight(BaseModel):
    """Pydantic class that defines each highlight to be returned by the LLM."""

    highlight: str = Field(
        ..., description=Prompts.descriptions["RESUME_SECTION_HIGHLIGHT"]["highlight"]
    )
    relevance: int = Field(
        ...,
        description=Prompts.descriptions["RESUME_SECTION_HIGHLIGHT"]["relevance"],
        enum=[1, 2, 3, 4, 5],
    )

class ResumeSectionHighlighterOutput(BaseModel):
    """Pydantic class that defines a list of highlights to be returned by the LLM."""

    plan: List[str] = Field(
        ...,
        description=Prompts.descriptions["RESUME_SECTION_HIGHLIGHTER_OUTPUT"]["plan"],
    )
    additional_steps: List[str] = Field(
        ...,
        description=Prompts.descriptions["RESUME_SECTION_HIGHLIGHTER_OUTPUT"]["additional_steps"],
    )
    work: List[str] = Field(
        ...,
        description=Prompts.descriptions["RESUME_SECTION_HIGHLIGHTER_OUTPUT"]["work"],
    )
    final_answer: List[ResumeSectionHighlight] = Field(
        ...,
        description=Prompts.descriptions["RESUME_SECTION_HIGHLIGHTER_OUTPUT"]["final_answer"],
    )

class ResumeSkills(BaseModel):
    """Pydantic class that defines a list of skills to be returned by the LLM."""

    technical_skills: List[str] = Field(
        ..., description=Prompts.descriptions["RESUME_SKILLS"]["technical_skills"]
    )
    non_technical_skills: List[str] = Field(
        ..., description=Prompts.descriptions["RESUME_SKILLS"]["non_technical_skills"]
    )

class ResumeSkillsMatcherOutput(BaseModel):
    """Pydantic class that defines a list of skills to be returned by the LLM."""

    plan: List[str] = Field(
        ..., description=Prompts.descriptions["RESUME_SKILLS_MATCHER_OUTPUT"]["plan"]
    )
    additional_steps: List[str] = Field(
        ...,
        description=Prompts.descriptions["RESUME_SKILLS_MATCHER_OUTPUT"]["additional_steps"],
    )
    work: List[str] = Field(
        ..., description=Prompts.descriptions["RESUME_SKILLS_MATCHER_OUTPUT"]["work"]
    )
    final_answer: ResumeSkills = Field(
        ...,
        description=Prompts.descriptions["RESUME_SKILLS_MATCHER_OUTPUT"]["final_answer"],
    )

class ResumeSummarizerOutput(BaseModel):
    """Pydantic class that defines a list of skills to be returned by the LLM."""

    plan: List[str] = Field(
        ..., description=Prompts.descriptions["RESUME_OBJECTIVE_OUTPUT"]["plan"]
    )
    additional_steps: List[str] = Field(
        ...,
        description=Prompts.descriptions["RESUME_OBJECTIVE_OUTPUT"]["additional_steps"],
    )
    work: List[str] = Field(
        ..., description=Prompts.descriptions["RESUME_OBJECTIVE_OUTPUT"]["work"]
    )
    final_answer: str = Field(
        ...,
        description=Prompts.descriptions["RESUME_OBJECTIVE_OUTPUT"]["final_answer"],
    )

class ResumeImprovements(BaseModel):
    """Pydantic class that defines a list of improvements to be returned by the LLM."""

    section: str = Field(
        ...,
        enum=[
            "objective",
            "education",
            "experience",
            "skills",
            "spelling and grammar",
            "other",
        ],
    )
    improvements: List[str] = Field(
        ..., description=Prompts.descriptions["RESUME_IMPROVEMENTS"]["improvements"]
    )

class ResumeImproverOutput(BaseModel):
    """Pydantic class that defines a list of improvements to be returned by the LLM."""

    plan: List[str] = Field(
        ..., description=Prompts.descriptions["RESUME_IMPROVER_OUTPUT"]["plan"]
    )
    additional_steps: List[str] = Field(
        ...,
        description=Prompts.descriptions["RESUME_IMPROVER_OUTPUT"]["additional_steps"],
    )
    work: List[str] = Field(
        ..., description=Prompts.descriptions["RESUME_IMPROVER_OUTPUT"]["work"]
    )
    final_answer: List[ResumeImprovements] = Field(
        ..., description=Prompts.descriptions["RESUME_IMPROVER_OUTPUT"]["final_answer"]
    )


import unittest
from ..utils.yaml_handler import read_yaml, write_yaml, dict_to_yaml_string
from ..utils.file_handler import (
    read_jobfile,
    generator_key_in_nested_dict,
    get_dict_field,
)
import os
from ..config import config


class TestYamlHandler(unittest.TestCase):
    def test_read_yaml(self):
        data = read_yaml(filename=config.DEFAULT_RESUME_PATH)
        self.assertIsInstance(data, dict)

    def test_write_yaml(self):
        data = {"key": "value"}
        write_yaml(
            data,
            filename=os.path.join(config.TESTS_DATA_PATH, "test_output.yaml"),
        )
        self.assertTrue(
            os.path.exists(
                os.path.join(config.TESTS_DATA_PATH, "test_output.yaml")
            )
        )

    def test_dict_to_yaml_string(self):
        data = {"key": "value"}
        yaml_string = dict_to_yaml_string(data)
        self.assertIn("key: value", yaml_string)


class TestFileHandler(unittest.TestCase):
    def test_generator_key_in_nested_dict(self):
        nested_dict = {"key1": {"key2": "value"}}
        result = list(generator_key_in_nested_dict("key2", nested_dict))
        self.assertIn("value", result)

    def test_get_dict_field(self):
        data_dict = {"field": "value"}
        result = get_dict_field("field", data_dict)
        self.assertEqual(result, "value")


if __name__ == "__main__":
    unittest.main()



import logging
import os
import configparser
from langchain_openai import ChatOpenAI

# Initialize logger
logger = logging.getLogger(__name__)

# Your resume filename here:
YOUR_RESUME_NAME = "sample_resume.yaml"

# Define project paths
PROJECT_PATH = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
DATA_PATH = os.path.join(PROJECT_PATH, "data")
TESTS_DATA_PATH = os.path.join(PROJECT_PATH, "tests/test_data/")
DEFAULT_RESUME_PATH = os.path.join(DATA_PATH, YOUR_RESUME_NAME)
BACKGROUND_TASKS_LOG = os.path.join(DATA_PATH, "background_tasks", "tasks.log")
RESOURCES_PATH = os.path.join(PROJECT_PATH, "resources")
PROMPTS_PATH = os.path.join(PROJECT_PATH, "prompts")
CONFIG_PATH = os.path.join(PROJECT_PATH, "config")
PROMPTS_YAML = os.path.join(PROMPTS_PATH, "prompts.yaml")
DESCRIPTIONS_YAML = os.path.join(PROMPTS_PATH, "extractor_descriptions.yaml")
REQUESTS_HEADERS = {
    "User-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/70.0.3538.102 Safari/537.36 Edge/18.19582"
}

# Define model configuration
CHAT_MODEL = ChatOpenAI
MODEL_NAME = "gpt-4o"
TEMPERATURE = 0.3
#OPEN_FILE_COMMAND = "cursor -r"
OPEN_FILE_COMMAND = "code -r"
MAX_CONCURRENT_WORKERS = 4
MAX_RETRIES = 3
BACKOFF_FACTOR = 5


# Confirm presence of OpenAI API key
def ensure_openai_api_key():
    if "OPENAI_API_KEY" not in os.environ:
        logger.info(
            "OPENAI_API_KEY not found in environment. User will be prompted to enter their key."
        )
        os.environ["OPENAI_API_KEY"] = input("Enter your OpenAI API key:")


ensure_openai_api_key()




import os
from .. import config

from reportlab.lib.enums import TA_CENTER, TA_RIGHT
from reportlab.lib.pagesizes import A4
from reportlab.lib.styles import ParagraphStyle
from reportlab.lib.units import inch
from reportlab.lib import colors
from reportlab.platypus import SimpleDocTemplate


def generate_doc_template(name, job_data_location):
    """
    Generate and return a SimpleDocTemplate for the resume PDF.

    Args:
        name (str): The name of the resume author.
        job_data_location (str): The path where the PDF will be saved.

    Returns:
        tuple: A tuple containing the document template for the resume PDF and the PDF location.
    """
    author_name_formatted = name.replace(" ", "_") + "_resume"
    pdf_location = os.path.join(job_data_location, f"{author_name_formatted}.pdf")
    doc = SimpleDocTemplate(
        pdf_location,
        pagesize=A4,
        showBoundary=0,
        leftMargin=0.1 * inch,
        rightMargin=0.1 * inch,
        topMargin=0.1 * inch,
        bottomMargin=0.1 * inch,
        title=f"{author_name_formatted}",
        author=name,
    )
    return (doc, pdf_location)


DEFAULT_PADDING = (1, 1)
DOCUMENT_ALIGNMENT = [
    ("ALIGN", (0, 0), (0, -1), "LEFT"),
    ("ALIGN", (1, 0), (1, -1), "RIGHT"),
    ("LEFTPADDING", (0, 0), (-1, -1), 0),
    ("RIGHTPADDING", (0, 0), (-1, -1), 0),
]
DEBUG_STYLE = ("GRID", (0, 0), (-1, -1), 0, colors.black)

PAGE_WIDTH, PAGE_HEIGHT = A4
FULL_COLUMN_WIDTH = PAGE_WIDTH - 1 * inch
JSON_PATH = os.path.join(config.DATA_PATH, "data.json")

FONT_PATHS = {
    "regular": os.path.join(config.RESOURCES_PATH, "fonts/calibri.ttf"),
    "bold": os.path.join(config.RESOURCES_PATH, "fonts/calibrib.ttf"),
    "italic": os.path.join(config.RESOURCES_PATH, "fonts/calibrii.ttf"),
}
FONT_NAMES = {
    "regular": "FONT_Regular",
    "bold": "FONT_Bold",
    "italic": "FONT_Italic",
}

PARAGRAPH_STYLES = {
    "bullet_points": ParagraphStyle(
        name="bullet_points_paragraph",
        leftIndent=12,
        fontName=FONT_NAMES["regular"],
        fontSize=11,
        leading=12,
        alignment=0,
        justifyBreaks=0,
        justifyLastLine=0,
    ),
    "last_bullet_point": ParagraphStyle(
        name="last_bullet_point",
        leftIndent=12,
        fontName=FONT_NAMES["regular"],
        fontSize=11,
        leading=12,
        alignment=0,
        justifyBreaks=0,
        justifyLastLine=0,
    ),
    "name": ParagraphStyle(
        name="name_paragraph",
        fontName=FONT_NAMES["bold"],
        fontSize=14,
        textTransform="uppercase",
        alignment=TA_CENTER,
        leading=12,
    ),
    "normal": ParagraphStyle(
        name="normal_paragraph",
        fontName=FONT_NAMES["regular"],
        fontSize=11,
        leading=12,
    ),
    "education": ParagraphStyle(
        name="education_paragraph",
        fontName=FONT_NAMES["regular"],
        fontSize=11,
        leading=12,
    ),
    "space": ParagraphStyle(
        name="space_paragraph",
        fontName=FONT_NAMES["regular"],
        fontSize=0,
        leading=0,
    ),
    "skills": ParagraphStyle(
        name="skills_paragraph",
        fontName=FONT_NAMES["regular"],
        fontSize=11,
        leading=12,
    ),
    "contact": ParagraphStyle(
        name="contact_paragraph",
        fontName=FONT_NAMES["regular"],
        fontSize=11,
        leading=12,
        alignment=TA_CENTER,
    ),
    "section": ParagraphStyle(
        name="section_paragraph",
        fontName=FONT_NAMES["bold"],
        fontSize=11,
        textTransform="uppercase",
    ),
    "objective": ParagraphStyle(
        name="objective_paragraph",
        fontName=FONT_NAMES["regular"],
        fontSize=11,
        leading=13,
        alignment=0,  # TA_LEFT
        justifyBreaks=0,
        justifyLastLine=0,
    ),
    "company_heading": ParagraphStyle(
        name="company_heading_paragraph",
        fontName=FONT_NAMES["bold"],
        fontSize=11,
        leading=10,
    ),
    "company_title": ParagraphStyle(
        name="company_title_paragraph",
        fontName=FONT_NAMES["italic"],
        fontSize=11,
        leading=10,
    ),
    "company_duration": ParagraphStyle(
        name="company_duration_paragraph",
        fontName=FONT_NAMES["italic"],
        fontSize=11,
        alignment=TA_RIGHT,
        leading=8,
    ),
    "company_location": ParagraphStyle(
        name="company_location_paragraph",
        fontName=FONT_NAMES["italic"],
        fontSize=11,
        alignment=TA_RIGHT,
        leading=8,
    ),
}



import configparser
import json
import os
import random
from .. import config
from .. import utils
import subprocess
import yaml

from reportlab.lib import colors
from reportlab.lib.enums import TA_CENTER, TA_RIGHT
from reportlab.lib.pagesizes import A4
from reportlab.lib.styles import ParagraphStyle
from reportlab.lib.units import inch
from reportlab.pdfbase import pdfmetrics, ttfonts
from reportlab.platypus import (
    Paragraph,
    SimpleDocTemplate,
    Table,
    TableStyle,
    Spacer,
    HRFlowable,
)
from . import resume_pdf_styles


class ResumePDFGenerator:
    """
    A class to generate a resume PDF from JSON data using the ReportLab library.
    """

    def __init__(self):
        """
        Initialize the ResumePDFGenerator by registering fonts.
        """
        self._register_fonts()

    def _register_fonts(self):
        """
        Register fonts for use in the PDF.
        """
        for style, path in resume_pdf_styles.FONT_PATHS.items():
            pdfmetrics.registerFont(
                ttfonts.TTFont(resume_pdf_styles.FONT_NAMES[style], path)
            )

    def _append_section_table_style(self, table_styles, row_index):
        """
        Append styles for section headers in the table.

        Args:
            table_styles (list): List of table styles to be extended.
            row_index (int): The current row index in the table.
        """
        table_styles.extend(
            [
                ("TOPPADDING", (0, row_index), (1, row_index), 5),
                ("BOTTOMPADDING", (0, row_index), (1, row_index), 2),
                ("LINEBELOW", (0, row_index), (-1, row_index), 0.1, colors.black),
            ]
        )

    def _add_table_row(
        self,
        table_data,
        table_styles,
        row_index,
        content_style_map,
        span=False,
        padding=resume_pdf_styles.DEFAULT_PADDING,
        bullet_point=None,
    ):
        """
        Add a row to the table data and update the table styles.

        Args:
            table_data (list): The table data to be extended.
            table_styles (list): The table styles to be extended.
            row_index (int): The current row index in the table.
            content_style_map (list of tuples): A list of tuples where each tuple contains content and its corresponding style.
            span (bool, optional): Whether the row should span across columns. Defaults to False.
            padding (tuple, optional): The top and bottom padding for the row. Defaults to (1, 1).
            bullet_point (str, optional): The bullet point to prepend to the content.
            spacer_height (int, optional): The height of the spacer to add after the row. Defaults to None.
        """
        if bullet_point:
            table_data.append(
                [
                    Paragraph(content, bulletText=bullet_point, style=style)
                    for content, style in content_style_map
                ]
            )
        else:
            table_data.append(
                [Paragraph(content, style) for content, style in content_style_map]
            )
        table_styles.extend(
            [
                ("BOTTOMPADDING", (0, row_index), (-1, row_index), padding[0]),
                ("TOPPADDING", (0, row_index), (-1, row_index), padding[1]),
            ]
        )
        if span:
            table_styles.append(("SPAN", (0, row_index), (1, row_index)))

        row_index += 1

        return row_index

    def add_experiences(self, table_data, table_styles, row_index, experiences):
        """
        Add experiences section to the resume.

        Args:
            table_data (list): The table data to be extended.
            table_styles (list): The table styles to be extended.
            row_index (int): The current row index in the table.
            experiences (list): The list of experiences to be added.
        """
        row_index = self._add_table_row(
            table_data=table_data,
            table_styles=table_styles,
            row_index=row_index,
            content_style_map=[
                ("Experience", resume_pdf_styles.PARAGRAPH_STYLES["section"])
            ],
            span=True,
        )
        self._append_section_table_style(table_styles, row_index - 1)

        for idx, job in enumerate(experiences):
            # Create a duration string from startdate and enddate
            duration = f"{job['titles'][0]['startdate']}-{job['titles'][0]['enddate']}"
            if job['skip_name']:
                row_index = self._add_table_row(
                    table_data=table_data,
                    table_styles=table_styles,
                    row_index=row_index,
                    content_style_map=[
                        (
                            job["titles"][0]["name"],
                            resume_pdf_styles.PARAGRAPH_STYLES["company_title"],
                        ),
                        (duration, resume_pdf_styles.PARAGRAPH_STYLES["company_duration"]),
                    ],
                )

            else:
                row_index = self._add_table_row(
                    table_data=table_data,
                    table_styles=table_styles,
                    row_index=row_index,
                    content_style_map=[
                        (
                            job["company"],
                            resume_pdf_styles.PARAGRAPH_STYLES["company_heading"],
                        ),
                        (duration, resume_pdf_styles.PARAGRAPH_STYLES["company_duration"]),
                    ],
                )

                row_index = self._add_table_row(
                    table_data=table_data,
                    table_styles=table_styles,
                    row_index=row_index,
                    content_style_map=[
                        (
                            job["titles"][0]["name"],
                            resume_pdf_styles.PARAGRAPH_STYLES["company_title"],
                        ),
                        (
                            job["location"],
                            resume_pdf_styles.PARAGRAPH_STYLES["company_location"],
                        ),
                    ],
                )

            for i, bullet_point in enumerate(job["highlights"]):
                style = (
                    resume_pdf_styles.PARAGRAPH_STYLES["last_bullet_point"]
                    if i == len(job["highlights"]) - 1
                    else resume_pdf_styles.PARAGRAPH_STYLES["bullet_points"]
                )
                padding = (0, 1)
                if i == len(job["highlights"]) - 1:
                    padding = (5, 1)
                row_index = self._add_table_row(
                    table_data=table_data,
                    table_styles=table_styles,
                    bullet_point="•",
                    row_index=row_index,
                    content_style_map=[
                        (
                            bullet_point,
                            style,
                        )
                    ],
                    span=True,
                    padding=padding,
                )

        return row_index

    def add_education(self, table_data, table_styles, row_index, education):
        """
        Add education section to the resume.

        Args:
            table_data (list): The table data to be extended.
            table_styles (list): The table styles to be extended.
            row_index (int): The current row index in the table.
            education (list): The list of education entries to be added.
        """
        row_index = self._add_table_row(
            table_data=table_data,
            table_styles=table_styles,
            row_index=row_index,
            content_style_map=[
                ("Education", resume_pdf_styles.PARAGRAPH_STYLES["section"])
            ],
            span=True,
        )
        self._append_section_table_style(table_styles, row_index - 1)

        for edu in education:
            degrees = ", ".join(edu["degrees"][0]["names"])
            row_index = self._add_table_row(
                table_data=table_data,
                table_styles=table_styles,
                row_index=row_index,
                content_style_map=[
                    (
                        f"<font name='{resume_pdf_styles.FONT_NAMES['bold']}'>{edu['school']}</font>, {degrees}",
                        resume_pdf_styles.PARAGRAPH_STYLES["education"],
                    )
                ],
                span=True,
            )
        return row_index

    def add_skills(self, table_data, table_styles, row_index, skills):
        """
        Add skills section to the resume.

        Args:
            table_data (list): The table data to be extended.
            table_styles (list): The table styles to be extended.
            row_index (int): The current row index in the table.
            skills (list): The list of skills to be added.
        """
        row_index = self._add_table_row(
            table_data=table_data,
            table_styles=table_styles,
            row_index=row_index,
            content_style_map=[
                ("Skills", resume_pdf_styles.PARAGRAPH_STYLES["section"])
            ],
            span=True,
        )
        self._append_section_table_style(table_styles, row_index - 1)

        for group in skills:
            group_keys = list(group.keys())
            skill_type = group[group_keys[0]]
            skills_list = group[group_keys[1]]
            skills_str = ", ".join(skills_list)
            formatted_skills = f"<font name='{resume_pdf_styles.FONT_NAMES['bold']}'>{skill_type}</font>: {skills_str}"
            row_index = self._add_table_row(
                table_data=table_data,
                table_styles=table_styles,
                row_index=row_index,
                content_style_map=[
                    (formatted_skills, resume_pdf_styles.PARAGRAPH_STYLES["skills"])
                ],
                span=True,
                padding=(2, 2),
            )
        return row_index

    def generate_resume(self, job_data_location, data):
        """
        Generate a resume PDF from JSON data.

        Args:
            job_data_location (str): The path where the PDF will be saved.
            data (dict): The JSON data containing resume information.
        """
        email = data["basic"]["email"]
        name = data["basic"]["name"]
        phone = data["basic"]["phone"]

        info = f"{email} | {phone}"
        def clean_url(url):
            return (
                url.replace("https://", "").replace("http://", "").replace("www.", "")
            )
        for website in data["basic"]["websites"]:
            info+=f" | {clean_url(website)}"

        address = data["basic"]["address"]
        info+= f" | {address}"
        doc, pdf_location = resume_pdf_styles.generate_doc_template(
            name, job_data_location
        )
        table_data = []
        table_styles = []
        row_index = 0

        if data.get("debug", False):
            table_styles.append(resume_pdf_styles.DEBUG_STYLE)

        table_styles.extend(resume_pdf_styles.DOCUMENT_ALIGNMENT)

        # Add name and contact information
        row_index = self._add_table_row(
            table_data=table_data,
            table_styles=table_styles,
            row_index=row_index,
            content_style_map=[(name, resume_pdf_styles.PARAGRAPH_STYLES["name"])],
            span=True,
        )

        row_index = self._add_table_row(
            table_data=table_data,
            table_styles=table_styles,
            row_index=row_index,
            content_style_map=[
                (
                    info,
                    resume_pdf_styles.PARAGRAPH_STYLES["contact"],
                )
            ],
            span=True,
        )

        # Add objective
        row_index = self._add_table_row(
            table_data=table_data,
            table_styles=table_styles,
            row_index=row_index,
            content_style_map=[
                ("Objective", resume_pdf_styles.PARAGRAPH_STYLES["section"])
            ],
            span=True,
        )
        self._append_section_table_style(table_styles, row_index - 1)

        row_index = self._add_table_row(
            table_data=table_data,
            table_styles=table_styles,
            row_index=row_index,
            content_style_map=[
                (data["objective"], resume_pdf_styles.PARAGRAPH_STYLES["objective"])
            ],
            span=True,
        )

        # Add experience
        row_index = self.add_experiences(
            table_data=table_data,
            table_styles=table_styles,
            row_index=row_index,
            experiences=data["experiences"],
        )

        # Add education
        row_index = self.add_education(
            table_data=table_data,
            table_styles=table_styles,
            row_index=row_index,
            education=data["education"],
        )

        # Add skills
        row_index = self.add_skills(
            table_data=table_data,
            table_styles=table_styles,
            row_index=row_index,
            skills=data["skills"],
        )

        table = Table(
            table_data,
            colWidths=[
                resume_pdf_styles.FULL_COLUMN_WIDTH * 0.75,
                resume_pdf_styles.FULL_COLUMN_WIDTH * 0.3,
            ],
            spaceBefore=0,
            spaceAfter=0,
        )
        table.setStyle(TableStyle(table_styles))

        doc.build([table])
        return pdf_location

    def generate_pdf_from_resume_yaml(self, yaml_path, job_data_location):
        """
        Generate a resume PDF from YAML data.

        Args:
            yaml_path (str): The path to the YAML file containing resume information.
            job_data_location (str): The path where the PDF will be saved.
        """
        return self.generate_resume(
            job_data_location, utils.read_yaml(filename=yaml_path)
        )



import threading
import time
from ..config import config
import logging


class BackgroundRunner:
    def __init__(self):
        self.threads = []
        self.logger = self._create_logger()

    def _create_logger(self):
        """Creates a custom logger for the BackgroundRunner."""
        logger = logging.getLogger(__name__)
        handler = logging.FileHandler(config.BACKGROUND_TASKS_LOG)
        formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(logging.INFO)
        return logger

    def run_in_background(self, func, *args, **kwargs):
        """Submits a function to be run in the background."""
        self.logger.info(
            f"Submitting task {func.__name__} with args {args} and kwargs {kwargs}"
        )
        thread = threading.Thread(target=func, args=args, kwargs=kwargs)
        self.threads.append(thread)
        thread.start()
        self.logger.info("Task submitted.")

    def check_status(self):
        """Checks the status of the background tasks."""
        if not self.threads:
            self.logger.info("No tasks submitted.")
            return "No tasks submitted."
        
        statuses = []
        for thread in self.threads:
            if thread.is_alive():
                self.logger.info("Task is running.")
                statuses.append("Task is running.")
            else:
                self.logger.info("Task completed.")
                statuses.append("Task completed.")
        return statuses

    def stop_all_tasks(self):
        """Stops all running background tasks."""
        self.logger.info("Stopping all tasks.")
        for thread in self.threads:
            if thread.is_alive():
                # Python does not provide a direct way to kill threads.
                self.logger.warning("Cannot directly stop threads. They will complete on their own.")
        self.logger.info("All tasks stopped.")




import yaml
from ruamel.yaml.error import YAMLError
from io import StringIO
from typing import Optional
from .. import config


def read_yaml(yaml_text: str = "", filename: str = "") -> Optional[dict]:
    """
    Reads YAML content from a string or a file.

    Args:
        yaml_text (str): YAML content as a string.
        filename (str): Path to the YAML file.

    Returns:
        dict: Parsed YAML content as a dictionary.
    """
    if not yaml_text and not filename:
        config.logger.warning("Neither yaml text nor filename have been provided.")
        return None
    if yaml_text:
        try:
            return yaml.load(yaml_text)
        except YAMLError as e:
            config.logger.error(f"The text could not be read.")
            raise e
    try:
        with open(filename, "r") as data:
            return yaml.safe_load(data)
    except YAMLError as e:
        config.logger.error(f"The {filename} could not be read.")
        raise e
    except Exception as e:
        config.logger.error(
            f"The {filename} could not be written due to an unknown error."
        )
        raise e


def write_yaml(data: dict, filename: str = None) -> None:
    """
    Writes a dictionary to a YAML file or prints it to stdout.

    Args:
        data (dict): Data to be written to YAML.
        filename (str): Path to the YAML file.
    """
    yaml.allow_unicode = True
    try:
        if filename:
            with open(filename, "w") as stream:
                yaml.dump(data, stream)
        else:
            yaml.dump(data, sys.stdout)
    except YAMLError as e:
        config.logger.error(f"The {filename} could not be written.")
        raise e
    except Exception as e:
        config.logger.error(
            f"The {filename} could not be written due to an unknown error."
        )
        raise e


def dict_to_yaml_string(data: dict) -> str:
    """
    Converts a dictionary to a YAML-formatted string.

    Args:
        data (dict): Data to be converted to YAML string.

    Returns:
        str: YAML-formatted string.
    """
    yaml.allow_unicode = True
    try:
        yaml.allow_unicode = True
        stream = StringIO()
        yaml.dump(data, stream=stream)
        return stream.getvalue()
    except YAMLError as e:
        config.logger.error("Failed to convert dict to YAML string.")
        raise e



import unittest
from ..services.resume_improver import ResumeImprover
from ..services.langchain_helpers import (
    create_llm,
    format_list_as_string,
    format_prompt_inputs_as_strings,
    parse_date,
    datediff_years,
)
from ..config import config


class TestResumeImproverExtractor(unittest.TestCase):
    def setUp(self):
        self.url = (
            "https://takline.github.io/ResumeGPT/tests/test_data/example_job_posting"
        )
        self.resume_improver = ResumeImprover(self.url)

    def test_download_and_parse_job_post(self):
        #self.resume_improver.download_and_parse_job_post()
        #with open(os.path.join(config.PROJECT_PATH, "tests/test_data/example_job_posting.html"), 'r', encoding='utf-8') as file:
        #    raw_html = file.read()
        #self.resume_improver.parse_raw_job_post(raw_html=raw_html)

        # Check HTML data extraction
        self.assertIn(
            "<title>Data Infrastructure Engineer | OpenAI | OpenAI</title>",
            self.resume_improver.job_post_html_data,
            "HTML title not found in job post data",
        )
        self.assertIn(
            "Data Infrastructure Engineer",
            self.resume_improver.job_post_raw,
            "Job title not found in raw job post data",
        )

        # Check parsed job data
        self.assertIsNotNone(self.resume_improver.parsed_job, "Parsed job data is None")
        self.assertIsInstance(
            self.resume_improver.parsed_job, dict, "Parsed job data is not a dictionary"
        )

        # Check specific fields in parsed job data
        self.assertIsNotNone(
            self.resume_improver.parsed_job.get("job_summary"), "Job summary is None"
        )
        self.assertGreater(
            len(self.resume_improver.parsed_job.get("qualifications", [])),
            0,
            "Qualifications list is empty",
        )
        self.assertGreater(
            len(self.resume_improver.parsed_job.get("duties", [])),
            0,
            "Duties list is empty",
        )
        self.assertGreater(
            len(self.resume_improver.parsed_job.get("ats_keywords", [])),
            0,
            "ATS keywords list is empty",
        )

        # Check specific values in parsed job data
        self.assertNotEqual(
            self.resume_improver.parsed_job.get("is_fully_remote"),
            True,
            "Job is incorrectly marked as fully remote",
        )
        self.assertIn(
            "$200K – $385K",
            self.resume_improver.parsed_job.get("salary", ""),
            "Expected salary range not found",
        )
        self.assertIn(
            "data infrastructure engineer",
            self.resume_improver.parsed_job.get("job_title", "").lower(),
            "Expected job title not found",
        )
        self.assertIn(
            "openai",
            self.resume_improver.parsed_job.get("company", "").lower(),
            "Expected company name not found",
        )
        self.assertIn(
            "python",
            ", ".join(self.resume_improver.parsed_job.get("ats_keywords", [])).lower(),
            "Expected ATS keyword 'python' not found",
        )
        self.assertIn(
            "4+ years in data infrastructure engineering",
            ", ".join(
                self.resume_improver.parsed_job.get("qualifications", [])
            ).lower(),
            "Expected qualification not found",
        )
        self.assertIn(
            "design, build, and maintain data infrastructure systems",
            ", ".join(self.resume_improver.parsed_job.get("duties", [])).lower(),
            "Expected duty not found",
        )

    def test_create_draft_tailored_resume(self):
        self.resume_improver.create_draft_tailored_resume(
            auto_open=False, manual_review=False, skip_pdf_create=True
        )
        self.assertTrue(self.resume_improver.yaml_loc.endswith("resume.yaml"))


class TestLangchainHelpers(unittest.TestCase):
    def test_create_llm(self):
        llm = create_llm()
        self.assertIsNotNone(llm)


if __name__ == "__main__":
    unittest.main()



import os
import subprocess
from jinja2 import Environment, FileSystemLoader
from ruamel.yaml import YAML
from ruamel.yaml.compat import StringIO
from ruamel.yaml.error import YAMLError
from .. import config
from . import yaml_handler


def generate_pdf(yaml_file: str, template_file: str = None) -> str:
    """
    Generates a PDF from a YAML file using a LaTeX template.

    Args:
        yaml_file (str): Path to the YAML file.
        template_file (str): Path to the LaTeX template file.

    Returns:
        str: Path to the generated PDF file.
    """
    # set default template file. file location is relative to `templates` directory
    if not template_file:
        template_file = "resume.tex"

    dirname, basename = os.path.split(yaml_file)
    filename, ext = os.path.splitext(basename)
    filename = os.path.join(dirname, filename)

    yaml_data = yaml_handler.read_yaml(filename=yaml_file)

    # Set up jinja environment and template
    env = Environment(
        trim_blocks=True,
        lstrip_blocks=True,
        block_start_string="\BLOCK{",
        block_end_string="}",
        variable_start_string="\VAR{",
        variable_end_string="}",
        comment_start_string="\#{",
        comment_end_string="}",
        line_statement_prefix="%%",
        line_comment_prefix="%#",
        autoescape=False,
        loader=FileSystemLoader("templates"),
    )
    template = env.get_template(template_file)

    latex_string = template.render(**yaml_data)

    try:
        with open(f"{filename}.tex", "wt") as stream:
            stream.write(latex_string)

        # convert to pdf and clean up temp files
        jobname = os.path.join(dirname, "latexmk_temp")
        cmd = subprocess.run(
            [
                "latexmk",
                "-pdf",
                f"-jobname={jobname}",
                f"{filename}.tex",
            ],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        # rename pdf
        if os.path.isfile(f"{jobname}.pdf"):
            os.rename(f"{jobname}.pdf", f"{filename}.pdf")
            cmd = subprocess.run(
                [
                    "latexmk",
                    "-c",
                    f"-jobname={jobname}",
                    f"{filename}.tex",
                ],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
        else:
            message = f"PDF could not be generated. See latexmk logs {jobname}.log"
            config.logger.error(message)
            raise ValueError(message)
    except Exception as e:
        config.logger.error(f"Failed to generate PDF: {e}")
        raise e

    return f"{filename}.pdf"



import os
from typing import Union, List, Generator, Optional
from .. import config
from .. import utils


def read_jobfile(filename: str) -> str:
    """
    Reads the content of a job file.

    Args:
        filename (str): Path to the job file.

    Returns:
        str: Content of the job file.
    """
    try:
        with open(filename, "r") as stream:
            return stream.read().strip()
    except OSError as e:
        config.logger.error(f"The {filename} could not be read.")
        raise e


def generator_key_in_nested_dict(
    keys: Union[str, List[str]], nested_dict: dict
) -> Generator:
    """
    Generates values for specified keys in a nested dictionary.

    Args:
        keys (Union[str, List[str]]): Key or list of keys to search for.
        nested_dict (dict): The nested dictionary to search in.

    Yields:
        Generator: Values corresponding to the specified keys.
    """
    if hasattr(nested_dict, "items"):
        for key, value in nested_dict.items():
            if (isinstance(keys, list) and key in keys) or key == keys:
                yield value
            if isinstance(value, dict):
                yield from utils.generator_key_in_nested_dict(keys, value)
            elif isinstance(value, list):
                for item in value:
                    yield from utils.generator_key_in_nested_dict(keys, item)


def get_dict_field(field: str, data_dict: dict) -> Optional[dict]:
    """
    Retrieves a field from a dictionary.

    Args:
        field (str): The field to retrieve.
        data_dict (dict): The dictionary to retrieve the field from.

    Returns:
        Optional[dict]: The value of the field, or None if the field is missing.
    """
    try:
        return data_dict[field]
    except KeyError as e:
        message = f"`{field}` is missing in raw resume."
        config.logger.warning(message)
    return None



from langchain_core.pydantic_v1 import BaseModel, Field
from typing import List, Optional
from ..prompts.prompts import Prompts
from ..services.extractor import ExtractorLLM

Prompts.initialize()

class JobDescription(BaseModel):
    """Description of a job posting."""

    company: Optional[str] = Field(
        None, description=Prompts.descriptions["JOB_DESCRIPTION"]["company"]
    )
    job_title: Optional[str] = Field(
        None, description=Prompts.descriptions["JOB_DESCRIPTION"]["job_title"]
    )
    team: Optional[str] = Field(
        None, description=Prompts.descriptions["JOB_DESCRIPTION"]["team"]
    )
    job_summary: Optional[str] = Field(
        None, description=Prompts.descriptions["JOB_DESCRIPTION"]["job_summary"]
    )
    salary: Optional[str] = Field(
        None, description=Prompts.descriptions["JOB_DESCRIPTION"]["salary"]
    )
    duties: Optional[List[str]] = Field(
        None, description=Prompts.descriptions["JOB_DESCRIPTION"]["duties"]
    )
    qualifications: Optional[List[str]] = Field(
        None, description=Prompts.descriptions["JOB_DESCRIPTION"]["qualifications"]
    )
    ats_keywords: Optional[List[str]] = Field(
        None, description=Prompts.descriptions["JOB_DESCRIPTION"]["ats_keywords"]
    )
    is_fully_remote: Optional[bool] = Field(
        None, description=Prompts.descriptions["JOB_DESCRIPTION"]["is_fully_remote"]
    )

class JobSkills(BaseModel):
    """Skills from a job posting."""

    technical_skills: Optional[List[str]] = Field(
        None, description=Prompts.descriptions["JOB_SKILLS"]["technical_skills"]
    )
    non_technical_skills: Optional[List[str]] = Field(
        None, description=Prompts.descriptions["JOB_SKILLS"]["non_technical_skills"]
    )

class JobPost(ExtractorLLM):
    def __init__(self, posting: str):
        """Initialize JobPost with the job posting string."""
        super().__init__()
        self.posting = posting
        self.parsed_job = None

    def parse_job_post(self, **chain_kwargs) -> dict:
        """Parse the job posting to extract job description and skills."""
        parsed_job = self.extract_from_input(
            pydantic_object=JobDescription, input=self.posting, **chain_kwargs
        )
        job_skills = self.extract_from_input(
            pydantic_object=JobSkills, input=self.posting, **chain_kwargs
        )
        self.parsed_job = {**parsed_job, **job_skills}
        return self.parsed_job



import unittest
from ..pdf_generation.resume_pdf_generator import ResumePDFGenerator
from ..config import config
from ..utils import utils
import os


class TestResumePDFGenerator(unittest.TestCase):
    def setUp(self):
        self.pdf_generator = ResumePDFGenerator()

    def test_generate_resume(self):
        data = utils.read_yaml(filename=config.DEFAULT_RESUME_PATH)
        file_path = os.path.join(config.PROJECT_PATH, "tests/test_data/")
        self.pdf_generator.generate_resume(job_data_location=file_path, data=data)
        self.assertTrue(os.path.exists(file_path))


if __name__ == "__main__":
    unittest.main()



import yaml
from .. import config

def check_resume_format(yaml_file_path: str) -> bool:
    """Check if the resume format is correct and provide suggestions for corrections.

    Args:
        yaml_file_path (str): The path to the resume YAML file.

    Returns:
        bool: True if the format is correct, False otherwise.
    """
    expected_format = {
        "editing": bool,
        "debug": bool,
        "basic": {
            "name": str,
            "address": str,
            "email": str,
            "phone": str,
            "websites": [str],
        },
        "objective": str,
        "education": [{"school": str, "degrees": [{"names": [str]}]}],
        "experiences": [
            {
                "company": str,
                "skip_name": bool,
                "location": str,
                "titles": [
                    {
                        "name": str,
                        "startdate": (int, str),  # startdate can be int or str
                        "enddate": (int, str),  # enddate can be int or str
                    }
                ],
                "highlights": [str],
            }
        ],
        "skills": [{"category": str, "skills": [str]}],
    }
    
    example_snippets = {
        "editing": "editing: true",
        "debug": "debug: false",
        "basic": """basic:
  name: John Doe
  address: Los Angeles, CA
  email: johndoe@example.com
  phone: 555-123-4567
  websites:
      - https://linkedin.com/johndoe
      - https://github.com/johndoe""",
        "objective": "objective: A Software Engineer with over 8 years of experience...",
        "education": """education:
  - school: University of California, Berkeley
    degrees:
      - names:
          - B.S. Computer Science
  - school: Stanford University
    degrees:
      - names:
          - M.S. Computer Science""",
        "experiences": """experiences:
  - company: Tech Innovators Inc.
    skip_name: false
    location: San Francisco, CA
    titles:
      - name: Lead Software Engineer
        startdate: 2022
        enddate: 2024
    highlights:
      - Led the development of a cloud-based platform, increasing user engagement by 50%.
      - Implemented a microservices architecture, reducing system downtime by 30%.
      - Mentored a team of junior developers, fostering a culture of continuous learning and improvement.
      - Spearheaded the integration of AI-driven features, enhancing product capabilities and user satisfaction.""",
        "skills": """skills:
  - category: Technical
    skills:
      - JavaScript
      - Python
      - AWS
      - Docker
      - Kubernetes
      - React
      - Node.js
      - Microservices
      - CI/CD
      - SQL
      - NoSQL
      - REST APIs
  - category: Non-technical
    skills:
      - Strong problem-solving skills
      - Excellent communication
      - Team leadership
      - Project management
      - Agile methodologies"""
    }
    
    def validate_format(actual, expected, path=""):
        """Recursively validate the format of the actual YAML content against the expected format.

        Args:
            actual (Any): The actual content from the YAML file.
            expected (Any): The expected format to validate against.
            path (str, optional): The current path in the YAML structure. Defaults to "".

        Returns:
            list: A list of errors found during validation.
        """
        if isinstance(expected, dict):
            if not isinstance(actual, dict):
                return [(path, "dict", type(actual).__name__)]
            errors = []
            for key, expected_type in expected.items():
                if key not in actual:
                    errors.append((f"{path}/{key}", expected_type, "missing"))
                else:
                    errors.extend(validate_format(actual[key], expected_type, f"{path}/{key}"))
            return errors
        elif isinstance(expected, list):
            if not isinstance(actual, list):
                return [(path, "list", type(actual).__name__)]
            if len(expected) == 0:
                return []  # No specific type to check against
            errors = []
            for index, item in enumerate(actual):
                errors.extend(validate_format(item, expected[0], f"{path}[{index}]"))
            return errors
        else:
            if not isinstance(actual, expected):
                return [(path, expected.__name__, type(actual).__name__)]
            return []

    with open(yaml_file_path, "r") as file:
        actual_yaml_dict = yaml.safe_load(file)

    errors = validate_format(actual_yaml_dict, expected_format)

    def get_example_snippet(path):
        """Retrieve an example snippet for a given path.

        Args:
            path (str): The path in the YAML structure.

        Returns:
            str: An example snippet corresponding to the path.
        """
        key = path.split('/')[1]
        key = key.split('[')[0]
        return example_snippets.get(key, None)

    consolidated_errors = {}

    for error in errors:
        path, expected, actual = error
        main_key = path.split('/')[1]
        sub_key = path.split('/')[-1]
        if main_key not in consolidated_errors:
            consolidated_errors[main_key] = {"missing": [], "incorrect": [], "entries": []}
        
        if main_key == "experiences":
            entry_index = path.split('/')[2].split('[')[-1][:-1]
            consolidated_errors[main_key]["entries"].append(f"experiences[{entry_index}]")
        
        if actual == "missing":
            consolidated_errors[main_key]["missing"].append(sub_key)
        elif isinstance(expected, str):
            consolidated_errors[main_key]["incorrect"].append((sub_key, actual, expected))
        else:
            consolidated_errors[main_key]["incorrect"].append((sub_key, actual, expected.__name__))
    logger_error = ""
    if consolidated_errors:
        for main_key, issues in consolidated_errors.items():
            example_snippet = get_example_snippet(f"/{main_key}")
            if main_key == "experiences" and issues["entries"]:
                entries = ", ".join(set(issues["entries"]))
                logger_error+=f"\nYou have formatting errors in these experiences entries: '{entries}'. Make sure they are formatted like this example:\n\n```yaml\n{example_snippet}\n```"
            if issues["missing"]:
                missing_keys = ", ".join(issues["missing"])
                logger_error+=f"\nYou are missing these keys: '{missing_keys}' in the '{main_key}' section. Make sure it is formatted like this example:\n\n```yaml\n{example_snippet}\n```"
            if issues["incorrect"]:
                for sub_key, actual_type, expected_type in issues["incorrect"]:
                    logger_error+=f"The value for '{sub_key}' in the '{main_key}' section is of type '{actual_type}'. \nExpected type: '{expected_type}'. Make sure it is formatted like this example:\n\n```yaml\n{example_snippet}\n```"
        config.logger.error(logger_error)
        return False
    else:
        return True



from setuptools import setup, find_packages

setup(
    name="ResumeGPT",
    version="2.0",
    description="A brief description of your project",
    author="Tyler Kline",
    author_email="tylerkline@gmail.com",
    url="https://github.com/takline/ResumeGPT",
    packages=find_packages(),
    install_requires=[
        "beautifulsoup4>=4.9.3",
        "configparser>=5.0.2",
        "langchain==0.1.20",
        "langchain-openai==0.1.6",
        "langchain-core==0.1.52",
        "pydantic==2.7.1",
        "reportlab>=3.5.59",
        "requests>=2.25.1",
        "ruamel.yaml>=0.16.12",
        "pytest>=8.2.2",
        "free-proxy>=1.1.1",
        "Jinja2==3.1.4",
        "PyYAML==5.3.1",
        "pytest-cov==5.0.0"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.11",
)



import os
import time
import subprocess
from datetime import datetime
from typing import List, Optional
from bs4 import BeautifulSoup
import uuid
import requests
from langchain.prompts import ChatPromptTemplate
from langchain_core.runnables import RunnableSequence
from langchain_core.output_parsers import StrOutputParser
from ..models.resume import (
    ResumeImproverOutput,
    ResumeSkillsMatcherOutput,
    ResumeSummarizerOutput,
    ResumeSectionHighlighterOutput,
)
from .. import utils
from .. import config
from .langchain_helpers import *
from .extractor import ExtractorLLM
from ..prompts import Prompts
from ..models.job_post import JobPost
from ..pdf_generation import ResumePDFGenerator
import concurrent.futures
from fp.fp import FreeProxy
import time
from ..config import config
from .background_runner import BackgroundRunner


class ResumeImprover(ExtractorLLM):

    def __init__(self, url, resume_location=None, llm_kwargs: dict = None):
        """Initialize ResumeImprover with the job post URL and optional resume location.

        Args:
            url (str): The URL of the job post.
            resume_location (str, optional): The file path to the resume. Defaults to None.
            llm_kwargs (dict, optional): Additional keyword arguments for the language model. Defaults to None.
        """
        super().__init__()
        self.job_post_html_data = None
        self.job_post_raw = None
        self.resume = None
        self.resume_yaml = None
        self.job_post = None
        self.parsed_job = None
        self.llm_kwargs = llm_kwargs or {}
        self.editing = False
        self.text_area = None
        self.clean_url = None
        self.job_data_location = None
        self.yaml_loc = None
        self.url = url
        self.download_and_parse_job_post()
        self.resume_location = resume_location or config.DEFAULT_RESUME_PATH
        self._update_resume_fields()

    def _update_resume_fields(self):
        """Update the resume fields based on the current resume location."""
        utils.check_resume_format(self.resume_location)
        self.resume = utils.read_yaml(filename=self.resume_location)
        self.degrees = self._get_degrees(self.resume)
        self.basic_info = utils.get_dict_field(field="basic", data_dict=self.resume)
        self.education = utils.get_dict_field(field="education", data_dict=self.resume)
        self.experiences = utils.get_dict_field(
            field="experiences", data_dict=self.resume
        )
        self.skills = utils.get_dict_field(field="skills", data_dict=self.resume)
        self.objective = utils.get_dict_field(field="objective", data_dict=self.resume)

    def update_resume(self, new_resume_location):
        """Update the resume location and refresh the dependent fields.

        Args:
            new_resume_location (str): The new file path to the resume.
        """
        self.resume_location = new_resume_location
        self._update_resume_fields()

    def _extract_html_data(self):
        """Extract text content from HTML, removing all HTML tags.

        Raises:
            Exception: If HTML data extraction fails.
        """
        try:
            soup = BeautifulSoup(self.job_post_html_data, "html.parser")
            self.job_post_raw = soup.get_text(separator=" ", strip=True)
        except Exception as e:
            config.logger.error(f"Failed to extract HTML data: {e}")
            raise

    def _download_url(self, url=None):
        """Download the content of the URL and return it as a string.

        Args:
            url (str, optional): The URL to download. Defaults to None.

        Returns:
            bool: True if download was successful, False otherwise.
        """
        if url:
            self.url = url

        max_retries = config.MAX_RETRIES
        backoff_factor = config.BACKOFF_FACTOR
        use_proxy = False

        for attempt in range(max_retries):
            try:
                proxies = None
                if use_proxy:
                    proxy = FreeProxy(rand=True).get()
                    proxies = {"http": proxy, "https": proxy}

                response = requests.get(
                    self.url, headers=config.REQUESTS_HEADERS, proxies=proxies
                )
                response.raise_for_status()
                self.job_post_html_data = response.text
                return True

            except requests.RequestException as e:
                if response.status_code == 429:
                    config.logger.warning(
                        f"Rate limit exceeded. Retrying in {backoff_factor * 2 ** attempt} seconds..."
                    )
                    time.sleep(backoff_factor * 2**attempt)
                    use_proxy = True
                else:
                    config.logger.error(f"Failed to download URL {self.url}: {e}")
                    return False

        config.logger.error(f"Exceeded maximum retries for URL {self.url}")
        return False

    def download_and_parse_job_post(self, url=None):
        """Download and parse the job post from the provided URL.

        Args:
            url (str, optional): The URL of the job post. Defaults to None.
        """
        if url:
            self.url = url
        self._download_url()
        self._extract_html_data()
        self.job_post = JobPost(self.job_post_raw)
        self.parsed_job = self.job_post.parse_job_post(verbose=False)
        try:
            filename = self.parsed_job["company"] + "_" + self.parsed_job["job_title"]
            filename = filename.replace(" ", "_")
        except KeyError:
            if "://" in self.url:
                filename = self.url.split("://")[1]
            else:
                filename = self.url
            url_paths = filename.split("/")
            filename = url_paths[0]
            if len(url_paths) > 1:
                filename = filename + "." + url_paths[-1]
        self.clean_url = filename
        filepath = os.path.join(config.DATA_PATH, self.clean_url)
        self.job_data_location = filepath
        os.makedirs(self.job_data_location, exist_ok=True)
        utils.write_yaml(
            self.parsed_job, filename=os.path.join(self.job_data_location, "job.yaml")
        )

    def parse_raw_job_post(self, raw_html):
        """Download and parse the job post from the provided URL.

        Args:
            url (str, optional): The URL of the job post. Defaults to None.
        """
        self.job_post_html_data = raw_html
        self._extract_html_data()
        self.job_post = JobPost(self.job_post_raw)
        self.parsed_job = self.job_post.parse_job_post(verbose=False)
        try:
            filename = self.parsed_job["company"] + "_" + self.parsed_job["job_title"]
            filename = filename.replace(" ", "_")
        except KeyError:
            if "://" in self.url:
                filename = self.url.split("://")[1]
            else:
                filename = self.url
            url_paths = filename.split("/")
            filename = url_paths[0]
            if len(url_paths) > 1:
                filename = filename + "." + url_paths[-1]
        self.clean_url = filename
        filepath = os.path.join(config.DATA_PATH, self.clean_url)
        self.job_data_location = filepath
        os.makedirs(self.job_data_location, exist_ok=True)
        utils.write_yaml(
            self.parsed_job, filename=os.path.join(self.job_data_location, "job.yaml")
        )

    def create_draft_tailored_resume(
        self, auto_open=True, manual_review=True, skip_pdf_create=False
    ):
        """Run a full review of the resume against the job post.

        Args:
            auto_open (bool, optional): Whether to automatically open the generated resume. Defaults to True.
            manual_review (bool, optional): Whether to wait for manual review. Defaults to True.
        """
        config.logger.info("Extracting matched skills...")
        self.skills = self.extract_matched_skills(verbose=False)
        config.logger.info("Writing objective...")
        self.objective = self.write_objective(verbose=False)
        config.logger.info("Updating bullet points...")
        self.experiences = self.rewrite_unedited_experiences(verbose=False)
        config.logger.info("Done updating...")
        self.yaml_loc = os.path.join(self.job_data_location, "resume.yaml")
        resume_dict = dict(
            editing=True,
            basic=self.basic_info,
            objective=self.objective,
            education=self.education,
            experiences=self.experiences,
            skills=self.skills,
        )
        utils.write_yaml(resume_dict, filename=self.yaml_loc)
        self.resume_yaml = utils.read_yaml(filename=self.yaml_loc)
        if auto_open:
            subprocess.run(config.OPEN_FILE_COMMAND.split(" ") + [self.yaml_loc])
        while manual_review and utils.read_yaml(filename=self.yaml_loc)["editing"]:
            time.sleep(5)
        config.logger.info("Saving PDF")
        if not skip_pdf_create:
            self.create_pdf(auto_open=auto_open)

    pass

    def _create_tailored_resume_in_background(
        self, auto_open=True, manual_review=True, background_runner=None
    ):
        """Run a full review of the resume against the job post.

        Args:
            auto_open (bool, optional): Whether to automatically open the generated resume. Defaults to True.
            manual_review (bool, optional): Whether to wait for manual review. Defaults to True.
        """
        if background_runner is not None:
            logger = background_runner.logger
        else:
            logger = config.logger
        logger.info("Extracting matched skills...")
        self.skills = self.extract_matched_skills(verbose=False)
        logger.info("Writing objective...")
        self.objective = self.write_objective(verbose=False)
        logger.info("Updating bullet points...")
        self.experiences = self.rewrite_unedited_experiences(verbose=False)
        logger.info("Done updating...")
        self.yaml_loc = os.path.join(self.job_data_location, "resume.yaml")
        resume_dict = dict(
            editing=True,
            basic=self.basic_info,
            objective=self.objective,
            education=self.education,
            experiences=self.experiences,
            skills=self.skills,
        )
        utils.write_yaml(resume_dict, filename=self.yaml_loc)
        self.resume_yaml = utils.read_yaml(filename=self.yaml_loc)

    def create_draft_tailored_resumes_in_background(background_configs: List[dict]):
        """Run 'create_draft_tailored_resume' for multiple configurations in the background.

        Args:
            background_configs (List[dict]): List of configurations for creating draft tailored resumes.
                Each configuration dictionary should have the following keys:
                - url (str): The URL of the job posting.
                - resume_location (str): The file path to the resume to be tailored.
                - auto_open (bool, optional): Whether to automatically open the generated resume. Defaults to True.
                - manual_review (bool, optional): Whether to wait for manual review. Defaults to True.
        """
        output = {}
        output["ResumeImprovers"] = []
        output["background_runner"] = BackgroundRunner()

        def run_config(background_config, resume_improver):
            try:
                resume_improver.download_and_parse_job_post()
                resume_improver._create_tailored_resume_in_background(
                    auto_open=background_config.get("auto_open", True),
                    manual_review=background_config.get("manual_review", True),
                )
            except Exception as e:
                output["background_runner"].logger.error(
                    f"An error occurred with config {config}: {e}"
                )

        for background_config in background_configs:
            output["ResumeImprovers"].append(
                ResumeImprover(
                    url=background_config["url"],
                    resume_location=background_config.get("resume_location"),
                )
            )
            output["background_runner"].run_in_background(
                run_config, background_config, output["ResumeImprovers"][-1]
            )
        return output

    def _section_highlighter_chain(self, **chain_kwargs) -> RunnableSequence:
        """Create a chain for highlighting relevant resume sections.

        Returns:
            RunnableSequence: The chain for highlighting resume sections.
        """
        prompt_msgs = Prompts.lookup["SECTION_HIGHLIGHTER"]
        prompt = ChatPromptTemplate(messages=prompt_msgs)
        llm = create_llm(**self.llm_kwargs)
        return prompt | llm | StrOutputParser()

    def _skills_matcher_chain(self, **chain_kwargs) -> RunnableSequence:
        """Create a chain for matching skills from the resume to the job posting.

        Returns:
            RunnableSequence: The chain for matching skills.
        """
        prompt_msgs = Prompts.lookup["SKILLS_MATCHER"]
        prompt = ChatPromptTemplate(messages=prompt_msgs)
        llm = create_llm(**self.llm_kwargs)
        return prompt | llm | StrOutputParser()

    def _objective_writer_chain(self, **chain_kwargs) -> RunnableSequence:
        """Create a chain for writing a compelling resume objective.

        Returns:
            RunnableSequence: The chain for writing the resume objective.
        """
        prompt_msgs = Prompts.lookup["OBJECTIVE_WRITER"]
        prompt = ChatPromptTemplate(messages=prompt_msgs)
        llm = create_llm(**self.llm_kwargs)
        return prompt | llm | StrOutputParser()

    def _improver_chain(self, **chain_kwargs) -> RunnableSequence:
        """Create a chain for critiquing and improving the resume.

        Returns:
            RunnableSequence: The chain for critiquing and improving the resume.
        """
        prompt_msgs = Prompts.lookup["IMPROVER"]
        prompt = ChatPromptTemplate(messages=prompt_msgs)
        llm = create_llm(**self.llm_kwargs)
        return prompt | llm | StrOutputParser()

    def _get_degrees(self, resume: dict):
        """Extract degrees from the resume.

        Args:
            resume (dict): The resume data.

        Returns:
            list: A list of degree names.
        """
        result = []
        for degrees in utils.generator_key_in_nested_dict("degrees", resume):
            for degree in degrees:
                if isinstance(degree["names"], list):
                    result.extend(degree["names"])
                elif isinstance(degree["names"], str):
                    result.append(degree["names"])
        return result

    def _format_skills_for_prompt(self, skills: list) -> list:
        """Format skills for inclusion in a prompt.

        Args:
            skills (list): The list of skills.

        Returns:
            list: A formatted list of skills.
        """
        result = []
        for cat in skills:
            curr = ""
            if cat.get("category", ""):
                curr += f"{cat['category']}: "
            if "skills" in cat:
                curr += "Proficient in "
                curr += ", ".join(cat["skills"])
                result.append(curr)
        return result

    def _get_cumulative_time_from_titles(self, titles) -> int:
        """Calculate the cumulative time from job titles.

        Args:
            titles (list): A list of job titles with start and end dates.

        Returns:
            int: The cumulative time in years.
        """
        result = 0.0
        for t in titles:
            if "startdate" in t and "enddate" in t:
                if t["enddate"] == "current":
                    last_date = datetime.today().strftime("%Y-%m-%d")
                else:
                    last_date = t["enddate"]
            result += datediff_years(start_date=t["startdate"], end_date=last_date)
        return round(result)

    def _format_experiences_for_prompt(self) -> list:
        """Format experiences for inclusion in a prompt.

        Returns:
            list: A formatted list of experiences.
        """
        result = []
        for exp in self.experiences:
            curr = ""
            if "titles" in exp:
                exp_time = self._get_cumulative_time_from_titles(exp["titles"])
                curr += f"{exp_time} years experience in:"
            if "highlights" in exp:
                curr += format_list_as_string(exp["highlights"], list_sep="\n  - ")
                curr += "\n"
                result.append(curr)
        return result

    def _combine_skills_in_category(self, l1: list[str], l2: list[str]):
        """Combine two lists of skills without duplicating lowercase entries.

        Args:
            l1 (list[str]): The first list of skills.
            l2 (list[str]): The second list of skills.
        """
        l1_lower = {i.lower() for i in l1}
        for i in l2:
            if i.lower() not in l1_lower:
                l1.append(i)

    def _combine_skill_lists(self, l1: list[dict], l2: list[dict]):
        """Combine two lists of skill categories without duplicating lowercase entries.

        Args:
            l1 (list[dict]): The first list of skill categories.
            l2 (list[dict]): The second list of skill categories.
        """
        l1_categories_lowercase = {s["category"].lower(): i for i, s in enumerate(l1)}
        for s in l2:
            if s["category"].lower() in l1_categories_lowercase:
                self._combine_skills_in_category(
                    l1[l1_categories_lowercase[s["category"].lower()]]["skills"],
                    s["skills"],
                )
            else:
                l1.append(s)

    def _print_debug_message(self, chain_kwargs: dict, chain_output_unformatted: str):
        """Print a debug message.

        Args:
            chain_kwargs (dict): The keyword arguments for the chain.
            chain_output_unformatted (str): The unformatted output from the chain.
        """
        message = "Final answer is missing from the chain output."
        return

    def rewrite_section(self, section: list | str, **chain_kwargs) -> dict:
        """Rewrite a section of the resume.

        Args:
            section (list | str): The section to rewrite.
            **chain_kwargs: Additional keyword arguments for the chain.

        Returns:
            dict: The rewritten section.
        """
        chain = self._section_highlighter_chain(**chain_kwargs)
        chain_inputs = format_prompt_inputs_as_strings(
            prompt_inputs=chain.input_schema().dict(),
            **self.parsed_job,
            degrees=self.degrees,
            experiences=utils.dict_to_yaml_string(dict(Experiences=self.experiences)),
            education=utils.dict_to_yaml_string(dict(Education=self.education)),
            skills=self._format_skills_for_prompt(self.skills),
            objective=self.objective,
        )
        chain_inputs["section"] = section
        section_revised_unformatted = chain.invoke(chain_inputs)
        section_revised = self.extract_from_input(
            pydantic_object=ResumeSectionHighlighterOutput,
            input=section_revised_unformatted,
        )
        section_revised = sorted(
            section_revised["final_answer"], key=lambda d: d["relevance"] * -1
        )
        return [s["highlight"] for s in section_revised]

    def rewrite_unedited_experiences(self, **chain_kwargs) -> dict:
        """Rewrite unedited experiences in the resume.

        Args:
            **chain_kwargs: Additional keyword arguments for the chain.

        Returns:
            dict: The rewritten experiences.
        """
        result = []
        for exp in self.experiences:
            exp = dict(exp)
            exp["highlights"] = self.rewrite_section(section=exp, **chain_kwargs)
            result.append(exp)
        return result

    def extract_matched_skills(self, **chain_kwargs) -> dict:
        """Extract matched skills from the resume and job post.

        Args:
            **chain_kwargs: Additional keyword arguments for the chain.

        Returns:
            dict: The extracted skills.
        """
        chain = self._skills_matcher_chain(**chain_kwargs)
        chain_inputs = format_prompt_inputs_as_strings(
            prompt_inputs=chain.input_schema().dict(),
            **self.parsed_job,
            degrees=self.degrees,
            experiences=utils.dict_to_yaml_string(dict(Experiences=self.experiences)),
            education=utils.dict_to_yaml_string(dict(Education=self.education)),
            skills=self._format_skills_for_prompt(self.skills),
            objective=self.objective,
        )
        extracted_skills_unformatted = chain.invoke(chain_inputs)
        extracted_skills = self.extract_from_input(
            pydantic_object=ResumeSkillsMatcherOutput,
            input=extracted_skills_unformatted,
        )
        if not extracted_skills or "final_answer" not in extracted_skills:
            return None
        extracted_skills = extracted_skills["final_answer"]
        result = []
        if "technical_skills" in extracted_skills:
            result.append(
                dict(category="Technical", skills=extracted_skills["technical_skills"])
            )
        if "non_technical_skills" in extracted_skills:
            result.append(
                dict(
                    category="Non-technical",
                    skills=extracted_skills["non_technical_skills"],
                )
            )
        self._combine_skill_lists(result, self.skills)
        return result

    def write_objective(self, **chain_kwargs) -> dict:
        """Write a objective for the resume.

        Args:
            **chain_kwargs: Additional keyword arguments for the chain.

        Returns:
            dict: The written objective.
        """
        chain = self._objective_writer_chain(**chain_kwargs)
        chain_inputs = format_prompt_inputs_as_strings(
            prompt_inputs=chain.input_schema().dict(),
            **self.parsed_job,
            degrees=self.degrees,
            experiences=utils.dict_to_yaml_string(dict(Experiences=self.experiences)),
            education=utils.dict_to_yaml_string(dict(Education=self.education)),
            skills=self._format_skills_for_prompt(self.skills),
            objective=self.objective,
        )
        objective_unformatted = chain.invoke(chain_inputs)
        objective = self.extract_from_input(
            pydantic_object=ResumeSummarizerOutput, input=objective_unformatted
        )
        if not objective or "final_answer" not in objective:
            return None
        return objective["final_answer"]

    def suggest_improvements(self, **chain_kwargs) -> dict:
        """Suggest improvements for the resume.

        Args:
            **chain_kwargs: Additional keyword arguments for the chain.

        Returns:
            dict: The suggested improvements.
        """
        chain = self._improver_chain(**chain_kwargs)
        chain_inputs = format_prompt_inputs_as_strings(
            prompt_inputs=chain.input_schema().dict(),
            **self.parsed_job,
            degrees=self.degrees,
            experiences=utils.dict_to_yaml_string(dict(Experiences=self.experiences)),
            education=utils.dict_to_yaml_string(dict(Education=self.education)),
            skills=self._format_skills_for_prompt(self.skills),
            objective=self.objective,
        )
        improvements_unformatted = chain.invoke(chain_inputs)
        improvements = self.extract_from_input(
            pydantic_object=ResumeImproverOutput, input=improvements_unformatted
        )
        if not improvements or "final_answer" not in improvements:
            return None
        return improvements["final_answer"]

    def finalize(self) -> dict:
        """Finalize the resume data.

        Returns:
            dict: The finalized resume data.
        """
        return dict(
            basic=self.basic_info,
            objective=self.objective,
            education=self.education,
            experiences=self.experiences,
            skills=self.skills,
        )

    def create_pdf(self, auto_open=True):
        """Create a PDF of the resume.

        Args:
            auto_open (bool, optional): Whether to automatically open the generated PDF. Defaults to True.

        Returns:
            str: The file path to the generated PDF.
        """
        pdf_generator = ResumePDFGenerator()
        pdf_location = pdf_generator.generate_resume(
            job_data_location=self.job_data_location,
            data=utils.read_yaml(filename=self.yaml_loc),
        )
        if auto_open:
            subprocess.run(config.OPEN_FILE_COMMAND.split(" ") + [pdf_location])
        return pdf_location



from datetime import datetime
from typing import List
from dateutil import parser as dateparser
from dateutil.relativedelta import relativedelta
from langchain_openai import ChatOpenAI
import langchain
from langchain_community.cache import InMemoryCache
from .. import config

# Set up LLM cache
langchain.llm_cache = InMemoryCache()


def create_llm(**kwargs):
    """Create an LLM instance with specified parameters."""
    chat_model = kwargs.pop("chat_model", ChatOpenAI)
    kwargs.setdefault("model_name", "gpt-4o")
    kwargs.setdefault("cache", False)
    return chat_model(**kwargs)


def format_list_as_string(lst: list, list_sep: str = "\n- ") -> str:
    """Format a list as a string with a specified separator."""
    if isinstance(lst, list):
        return list_sep + list_sep.join(lst)
    return str(lst)


def format_prompt_inputs_as_strings(prompt_inputs: list[str], **kwargs):
    """Convert values to string for all keys in kwargs matching list in prompt inputs."""
    return {
        k: format_list_as_string(v) for k, v in kwargs.items() if k in prompt_inputs
    }


def parse_date(date_str: str) -> datetime:
    """Given an arbitrary string, parse it to a date."""
    default_date = datetime(datetime.today().year, 1, 1)
    try:
        return dateparser.parse(str(date_str), default=default_date)
    except dateparser._parser.ParserError as e:
        langchain.llm_cache.clear()
        config.logger.error(f"Date input `{date_str}` could not be parsed.")
        raise e


def datediff_years(start_date: str, end_date: str) -> float:
    """Get difference between arbitrarily formatted dates in fractional years to the floor month."""
    datediff = relativedelta(parse_date(end_date), parse_date(start_date))
    return datediff.years + datediff.months / 12.0



from langchain.prompts import ChatPromptTemplate, HumanMessagePromptTemplate
from langchain.schema import HumanMessage, SystemMessage
import yaml
from .. import config


class Prompts:
    """
    A class to load and manage prompt templates and extractor descriptions from a YAML configuration file.
    """

    lookup = None
    descriptions = None

    @classmethod
    def initialize(cls):
        """
        Initialize the Prompts class by loading the YAML files and setting up the lookup dictionary.
        """
        cls.lookup = cls._load_prompts(config.PROMPTS_YAML)
        cls.descriptions = cls._load_descriptions(config.DESCRIPTIONS_YAML)

    @staticmethod
    def _load_prompts(yaml_path: str) -> dict:
        """
        Load prompts from a YAML file and organize them into a lookup dictionary.

        :param yaml_path: Path to the YAML file containing prompt configurations.
        :return: A dictionary with prompt types as keys and lists of message templates as values.
        """
        with open(yaml_path, "r") as file:
            prompts_data = yaml.safe_load(file)

        lookup = {}
        for prompt_type, sub_data in prompts_data.items():
            sub_lookup = [
                SystemMessage(content=sub_data["system_message"]),
                HumanMessagePromptTemplate.from_template(
                    sub_data["job_posting_template"]
                ),
                HumanMessagePromptTemplate.from_template(
                    sub_data.get("master_resume_template", "")
                ),
                HumanMessage(content=sub_data["instruction_message"]),
                HumanMessage(content=sub_data["criteria_message"]),
                HumanMessage(content=sub_data["steps_message"]),
            ]
            lookup[prompt_type] = sub_lookup

        return lookup

    @staticmethod
    def _load_descriptions(yaml_path: str) -> dict:
        """
        Load descriptions from a YAML file.

        :param yaml_path: Path to the YAML file containing descriptions.
        :return: A dictionary with descriptions.
        """
        with open(yaml_path, "r") as file:
            descriptions_data = yaml.safe_load(file)
        return descriptions_data



import unittest
from ..models.job_post import JobDescription, JobSkills, JobPost
from ..models.resume import ResumeSectionHighlight, ResumeSectionHighlighterOutput, ResumeSkills, ResumeSkillsMatcherOutput, ResumeSummarizerOutput, ResumeImprovements, ResumeImproverOutput

class TestJobDescription(unittest.TestCase):
    def test_job_description_fields(self):
        job_description = JobDescription(company="Example Corp", job_title="Software Engineer")
        self.assertEqual(job_description.company, "Example Corp")
        self.assertEqual(job_description.job_title, "Software Engineer")

class TestJobSkills(unittest.TestCase):
    def test_job_skills_fields(self):
        job_skills = JobSkills(technical_skills=["Python"], non_technical_skills=["Communication"])
        self.assertIn("Python", job_skills.technical_skills)
        self.assertIn("Communication", job_skills.non_technical_skills)

class TestResumeModels(unittest.TestCase):
    def test_resume_section_highlight(self):
        highlight = ResumeSectionHighlight(highlight="Led a team", relevance=5)
        self.assertEqual(highlight.highlight, "Led a team")
        self.assertEqual(highlight.relevance, 5)

    def test_resume_skills(self):
        skills = ResumeSkills(technical_skills=["Python"], non_technical_skills=["Communication"])
        self.assertIn("Python", skills.technical_skills)
        self.assertIn("Communication", skills.non_technical_skills)

if __name__ == "__main__":
    unittest.main()
